kernel void nbody(global float* p, global float* pNeu, global float* v, const float DELTA_T, const float EPSILON_SQUARED)  
{
    uint id = get_global_id(0);
    uint N = get_global_size(0);
    
    uint elems = 3;

    float px = p[elems * id + 0];
    float py = p[elems * id + 1];
    float pz = p[elems * id + 2];
    
    float ax = 0;
    float ay = 0;
    float az = 0; 
    
    for(uint i = 0; i < N; ++i)
    {  
        float opx = p[elems * i + 0];
        float opy = p[elems * i + 1];
        float opz = p[elems * i + 2];
    
        float rx = opx - px;
        float ry = opy - py;
        float rz = opz - pz;

        float distSqr = rx * rx + ry * ry + rz * rz + EPSILON_SQUARED;  

        float invDist = rsqrt(distSqr);
        float invDistCube = invDist * invDist * invDist;

        ax += rx * invDistCube;
        ay += ry * invDistCube;  
        az += rz * invDistCube;  
    }

    float vx = v[elems * id + 0];
    float vy = v[elems * id + 1];
    float vz = v[elems * id + 2];
    
    vx += ax * DELTA_T;
    vy += ay * DELTA_T;
    vz += az * DELTA_T;
    
    px += vx * DELTA_T;
    py += vy * DELTA_T;
    pz += vz * DELTA_T;
    
    v[elems * id + 0] = vx;
    v[elems * id + 1] = vy;
    v[elems * id + 2] = vz;
    
    pNeu[elems * id + 0] = px;
    pNeu[elems * id + 1] = py;
    pNeu[elems * id + 2] = pz;
}

float4 bodyBodyInteraction(float4 pi, float4 pj, float4 ai, float EPSILON_SQUARED)  
{  
    float3 r;  

    r.x = pj.x - pi.x;  
    r.y = pj.y - pi.y;  
    r.z = pj.z - pi.z;  

    float distSqr = r.x * r.x + r.y * r.y + r.z * r.z + EPSILON_SQUARED;  

    float invDist = rsqrt(distSqr);
    float invDistCube =  invDist * invDist * invDist;
    //mass = 1
    float m = 1;
    float s = m * invDistCube;  

    ai.x += r.x * s;  
    ai.y += r.y * s;  
    ai.z += r.z * s;  
    return ai;  
}

kernel void nBody_CalcNewV(
 global float4* body_Pos,
 global float4* body_V,
 const float DELTA_T,
 const float EPSILON_SQUARED)
{
    uint id = get_global_id(0);
    uint N = get_global_size(0);

    float4 myPos = body_Pos[id];
    
    float4 ac = (float4)(0,0,0,0);
    
    for(uint i = 0; i < N; ++i)
    {  
        float4 otherPos = body_Pos[i];
        ac = bodyBodyInteraction(myPos, otherPos, ac, EPSILON_SQUARED);
    }
    
    float4 v = body_V[id];
    v += ac * DELTA_T;
    body_V[id] = v;
}

kernel void nBody_CalcNewPos(
 global float4* body_Pos, 
 global float4* body_V,
 const float DELTA_T)
{
    uint id = get_global_id(0);
    body_Pos[id] += DELTA_T * body_V[id];
}

kernel void passPositionOn(
 global float4* body_Pos,
 global float4* curveVertex_Pos,
 const uint VERTICES_PER_CURVE)
{
    uint id = get_global_id(0);
    uint startPos = id * VERTICES_PER_CURVE;   
    
    for(int i = 0; i < VERTICES_PER_CURVE - 1; i++)
        curveVertex_Pos[startPos + VERTICES_PER_CURVE - i - 1] =  curveVertex_Pos[startPos + VERTICES_PER_CURVE - i - 2];
    curveVertex_Pos[startPos] = body_Pos[id];
    
}

kernel void setTrailParticle(
   global float4* trailParticle_Pos, 
   global float*  trailParticle_S, 
   global float4* trailParticle_Dir, 
   global float4* curveVertex_Pos,   
   const int TRAIL_PARTICLES_PER_CURVE,
   const int VERTICES_PER_CURVE,   
   const float dSC,
   const float dSTP)
 {	
	int particle = get_global_id(0);
	int curve = particle / TRAIL_PARTICLES_PER_CURVE;
	int firstCV = curve * VERTICES_PER_CURVE;	

	float sTP = trailParticle_S[particle];
   	
    int indexPrevCV  = (int) (sTP * (VERTICES_PER_CURVE - 1));
    int indexNextCV = min(indexPrevCV  + 1, VERTICES_PER_CURVE - 1);
    float sPrevCV = indexPrevCV  * dSC;  
    float sTPLocal  = (sTP - sPrevCV) / dSC; 
    float4 ndir = curveVertex_Pos[firstCV + indexNextCV] - curveVertex_Pos[firstCV + indexPrevCV];
    
    ndir.w = sTP;
    
    trailParticle_Dir[particle] = ndir;
    
    float4 npos = sTPLocal * curveVertex_Pos[firstCV + indexNextCV] + (1-sTPLocal) * curveVertex_Pos[firstCV + indexPrevCV];
    
    trailParticle_Pos[particle] = npos;
		
	sTP += dSTP;
	if(sTP > 1)
		sTP -= 1;	
	trailParticle_S[particle] = sTP;
}